# -*- coding: utf-8 -*-
"""
BaseWorker methods for training.
"""

import hashlib
import os
import tarfile
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import NewType, Optional, Tuple

import requests
import zstandard as zstd
from apistar.exceptions import ErrorResponse

from arkindex_worker import logger

CHUNK_SIZE = 1024

DirPath = NewType("DirPath", str)
"""Path to a directory"""

Hash = NewType("Hash", str)
"""MD5 Hash"""

FileSize = NewType("FileSize", int)
"""File size"""


@contextmanager
def create_archive(path: DirPath) -> Tuple[Path, Hash, FileSize, Hash]:
    """
    Create a tar archive from the files at the given location then compress it to a zst archive.

    Yield its location, its hash, its size and its content's hash.

    :param path: Create a compressed tar archive from the files
    :returns: The location of the created archive, its hash, its size and its content's hash
    """
    assert path.is_dir(), "create_archive needs a directory"

    compressor = zstd.ZstdCompressor(level=3)
    content_hasher = hashlib.md5()
    archive_hasher = hashlib.md5()

    # Remove extension from the model filename
    _, path_to_tar_archive = tempfile.mkstemp(prefix="teklia-", suffix=".tar")

    # Create an uncompressed tar archive with all the needed files
    # Files hierarchy ifs kept in the archive.
    file_list = []
    with tarfile.open(path_to_tar_archive, "w") as tar:
        for p in path.glob("**/*"):
            x = p.relative_to(path)
            tar.add(p, arcname=x, recursive=False)
            file_list.append(p)

    # Sort by path
    file_list.sort()
    # Compute hash of the files
    for file_path in file_list:
        with open(file_path, "rb") as file_data:
            for chunk in iter(lambda: file_data.read(CHUNK_SIZE), b""):
                content_hasher.update(chunk)

    _, path_to_zst_archive = tempfile.mkstemp(prefix="teklia-", suffix=".tar.zst")

    # Compress the archive
    with open(path_to_zst_archive, "wb") as archive_file:
        with open(path_to_tar_archive, "rb") as model_data:
            for model_chunk in iter(lambda: model_data.read(CHUNK_SIZE), b""):
                compressed_chunk = compressor.compress(model_chunk)
                archive_hasher.update(compressed_chunk)
                archive_file.write(compressed_chunk)

    # Remove the tar archive
    os.remove(path_to_tar_archive)

    # Get content hash, archive size and hash
    hash = content_hasher.hexdigest()
    size = os.path.getsize(path_to_zst_archive)
    archive_hash = archive_hasher.hexdigest()

    yield path_to_zst_archive, hash, size, archive_hash

    # Remove the zstd archive
    os.remove(path_to_zst_archive)


class TrainingMixin(object):
    def publish_model_version(
        self,
        model_path: DirPath,
        model_id: str,
        tag: Optional[str] = None,
        description: Optional[str] = None,
    ):
        """
        This method creates a model archive and its associated hash,
        to create a unique version that will be stored on a bucket and published on Arkindex.

        :param model_path: Path to the directory containing the model version's files.
        :param model_id: ID of the model
        :param tag:
        :param description:
        """

        if self.is_read_only:
            logger.warning(
                "Cannot publish a new model version as this worker is in read-only mode"
            )
            return

        # Create the zst archive, get its hash and size
        with create_archive(path=model_path) as (
            path_to_archive,
            hash,
            size,
            archive_hash,
        ):
            # Create a new model version with hash and size
            model_version_details = self.create_model_version(
                model_id=model_id,
                hash=hash,
                size=size,
                archive_hash=archive_hash,
                tag=tag,
                description=description,
            )
            if model_version_details is None:
                return
            self.upload_to_s3(
                archive_path=path_to_archive,
                model_version_details=model_version_details,
            )

        # Update the model version with state, configuration parsed, tag, description (defaults to name of the worker)
        self.update_model_version(
            model_version_details=model_version_details,
        )

    def create_model_version(
        self,
        model_id: str,
        hash: str,
        size: int,
        archive_hash: str,
        tag: str,
        description: str,
    ) -> dict:
        """
        Create a new version of the specified model with the given information (hashes and size).
        If a version matching the information already exist, there are two cases:
        - The version is in `Created` state: this version's details is used
        - The version is in `Available` state: you cannot create twice the same version, an error is raised
        """
        if self.is_read_only:
            logger.warning(
                "Cannot create a new model version as this worker is in read-only mode"
            )
            return

        # Create a new model version with hash and size
        try:
            payload = {"hash": hash, "size": size, "archive_hash": archive_hash}
            if tag:
                payload["tag"] = tag
            if description:
                payload["description"] = description
            model_version_details = self.request(
                "CreateModelVersion",
                id=model_id,
                body=payload,
            )
            logger.info(
                f"Model version ({model_version_details['id']}) was created successfully"
            )
        except ErrorResponse as e:
            model_version_details = (
                e.content.get("hash") if hasattr(e, "content") else None
            )
            if e.status_code >= 500 or model_version_details is None:
                logger.error(f"Failed to create model version: {e.content}")
                raise e
            # If the existing model is in Created state, this model is returned as a dict.
            # Else an error in a list is returned.
            if isinstance(model_version_details, (list, tuple)):
                logger.error(model_version_details[0])
                return

            logger.info(
                f"Model version ({model_version_details['id']}) has the same hash, using this one instead of creating one"
            )

        return model_version_details

    def upload_to_s3(self, archive_path: str, model_version_details: dict) -> None:
        """
        Upload the archive of the model's files to an Amazon s3 compatible storage
        """
        if self.is_read_only:
            logger.warning(
                "Cannot upload this archive as this worker is in read-only mode"
            )
            return

        s3_put_url = model_version_details.get("s3_put_url")
        logger.info("Uploading to s3...")
        # Upload the archive on s3
        with open(archive_path, "rb") as archive:
            r = requests.put(
                url=s3_put_url,
                data=archive,
                headers={"Content-Type": "application/zstd"},
            )
        r.raise_for_status()

    def update_model_version(
        self,
        model_version_details: dict,
        configuration: dict = {},
    ) -> None:
        """
        Update the specified model version to the state `Available` and use the given information"
        """
        if self.is_read_only:
            logger.warning(
                "Cannot update this model version as this worker is in read-only mode"
            )
            return

        model_version_id = model_version_details.get("id")
        logger.info(f"Updating model version ({model_version_id})")
        try:
            self.request(
                "UpdateModelVersion",
                id=model_version_id,
                body={
                    "state": "available",
                    "description": model_version_details.get("description"),
                    "configuration": configuration,
                    "tag": model_version_details.get("tag"),
                },
            )
            logger.info(f"Model version ({model_version_id}) was successfully updated")
        except ErrorResponse as e:
            logger.error(f"Failed to update model version: {e.content}")



# -*- coding: utf-8 -*-
import os
import sys

import pytest
import responses

from arkindex.mock import MockApiClient
from arkindex_worker.worker import BaseWorker
from arkindex_worker.worker.training import TrainingMixin, create_archive


class TrainingWorker(BaseWorker, TrainingMixin):
    """
    This class is only needed for tests
    """

    pass


@pytest.fixture
def mock_training_worker(monkeypatch):
    monkeypatch.setattr(sys, "argv", ["worker"])
    training_worker = TrainingWorker()
    training_worker.api_client = MockApiClient()
    training_worker.args = training_worker.parser.parse_args()
    return training_worker


def test_create_archive(model_file_dir):
    """Create an archive when the model's file is in a folder"""

    with create_archive(path=model_file_dir) as (
        zst_archive_path,
        hash,
        size,
        archive_hash,
    ):
        assert os.path.exists(zst_archive_path), "The archive was not created"
        assert (
            hash == "c5aedde18a768757351068b840c8c8f9"
        ), "Hash was not properly computed"
        assert 300 < size < 700

    assert not os.path.exists(zst_archive_path), "Auto removal failed"


@pytest.mark.parametrize(
    "tag, description",
    [
        ("tag", "description"),
        (None, "description"),
        ("", "description"),
        ("tag", ""),
        ("", ""),
        (None, None),
    ],
)
def test_create_model_version(mock_training_worker, tag, description):
    """A new model version is returned"""

    model_version_id = "fake_model_version_id"
    model_id = "fake_model_id"
    model_hash = "hash"
    archive_hash = "archive_hash"
    size = "30"
    model_version_details = {
        "id": model_version_id,
        "model_id": model_id,
        "hash": model_hash,
        "archive_hash": archive_hash,
        "size": size,
        "tag": tag,
        "description": description,
        "s3_url": "http://hehehe.com",
        "s3_put_url": "http://hehehe.com",
    }

    expected_payload = {
        "hash": model_hash,
        "archive_hash": archive_hash,
        "size": size,
    }
    if description:
        expected_payload["description"] = description
    if tag:
        expected_payload["tag"] = tag

    mock_training_worker.api_client.add_response(
        "CreateModelVersion",
        id=model_id,
        response=model_version_details,
        body=expected_payload,
    )
    assert (
        mock_training_worker.create_model_version(
            model_id, model_hash, size, archive_hash, tag, description
        )
        == model_version_details
    )


@pytest.mark.parametrize(
    "content, status_code",
    [
        (
            {
                "id": "fake_model_version_id",
                "model_id": "fake_model_id",
                "hash": "hash",
                "archive_hash": "archive_hash",
                "size": "size",
                "tag": "tag",
                "description": "description",
                "s3_url": "http://hehehe.com",
                "s3_put_url": "http://hehehe.com",
            },
            400,
        ),
        (["A version for this model with this hash already exists."], 403),
    ],
)
def test_retrieve_created_model_version(mock_training_worker, content, status_code):
    """
    If there is an existing model version in Created mode,
    A 400 was raised, but the model is still returned in error content.
    Else if an existing model version in Available mode,
    403 was raised, but None will be returned
    """
    model_id = "fake_model_id"
    model_hash = "hash"
    archive_hash = "archive_hash"
    size = "30"
    mock_training_worker.api_client.add_error_response(
        "CreateModelVersion",
        id=model_id,
        status_code=status_code,
        body={
            "hash": model_hash,
            "archive_hash": archive_hash,
            "size": size,
        },
        content={"hash": content},
    )
    if status_code == 400:
        assert (
            mock_training_worker.create_model_version(
                model_id, model_hash, size, archive_hash, tag=None, description=None
            )
            == content
        )
    elif status_code == 403:
        assert (
            mock_training_worker.create_model_version(
                model_id, model_hash, size, archive_hash, tag=None, description=None
            )
            is None
        )


@pytest.mark.parametrize(
    "content, status_code",
    (
        # error 500
        ({"id": "fake_id"}, 500),
        # model_version details is None
        ({}, 403),
        (None, 403),
    ),
)
def test_handle_500_create_model_version(mock_training_worker, content, status_code):
    model_id = "fake_model_id"
    model_hash = "hash"
    archive_hash = "archive_hash"
    size = "30"
    mock_training_worker.api_client.add_error_response(
        "CreateModelVersion",
        id=model_id,
        status_code=status_code,
        body={
            "hash": model_hash,
            "archive_hash": archive_hash,
            "size": size,
        },
        content=content,
    )
    with pytest.raises(Exception):
        mock_training_worker.create_model_version(
            model_id, model_hash, size, archive_hash, tag=None, description=None
        )


def test_handle_s3_uploading_errors(mock_training_worker, model_file_dir):
    s3_endpoint_url = "http://s3.localhost.com"
    responses.add_passthru(s3_endpoint_url)
    responses.add(responses.Response(method="PUT", url=s3_endpoint_url, status=400))
    file_path = model_file_dir / "model_file.pth"
    with pytest.raises(Exception):
        mock_training_worker.upload_to_s3(file_path, {"s3_put_url": s3_endpoint_url})



# -*- coding: utf-8 -*-
"""
ElementsWorker methods for entities.
"""

from enum import Enum
from typing import Dict, Optional, Union

from peewee import IntegrityError

from arkindex_worker import logger
from arkindex_worker.cache import CachedElement, CachedEntity, CachedTranscriptionEntity
from arkindex_worker.models import Element, Transcription


class EntityType(Enum):
    """
    Type of an entity.
    """

    Person = "person"
    Location = "location"
    Subject = "subject"
    Organization = "organization"
    Misc = "misc"
    Number = "number"
    Date = "date"


class EntityMixin(object):
    def create_entity(
        self,
        element: Union[Element, CachedElement],
        name: str,
        type: EntityType,
        metas=dict(),
        validated=None,
    ):
        """
        Create an entity on the given corpus.
        If cache support is enabled, a [CachedEntity][arkindex_worker.cache.CachedEntity] will also be created.

        :param element: An element on which the entity will be reported with the [Reporter][arkindex_worker.reporting.Reporter].
           This does not have any effect on the entity itself.
        :param name: Name of the entity.
        :param type: Type of the entity.
        """
        assert element and isinstance(
            element, (Element, CachedElement)
        ), "element shouldn't be null and should be an Element or CachedElement"
        assert name and isinstance(
            name, str
        ), "name shouldn't be null and should be of type str"
        assert type and isinstance(
            type, EntityType
        ), "type shouldn't be null and should be of type EntityType"
        if metas:
            assert isinstance(metas, dict), "metas should be of type dict"
        if validated is not None:
            assert isinstance(validated, bool), "validated should be of type bool"
        if self.is_read_only:
            logger.warning("Cannot create entity as this worker is in read-only mode")
            return

        entity = self.request(
            "CreateEntity",
            body={
                "name": name,
                "type": type.value,
                "metas": metas,
                "validated": validated,
                "corpus": self.corpus_id,
                "worker_version": self.worker_version_id,
            },
        )
        self.report.add_entity(element.id, entity["id"], type.value, name)

        if self.use_cache:
            # Store entity in local cache
            try:
                to_insert = [
                    {
                        "id": entity["id"],
                        "type": type.value,
                        "name": name,
                        "validated": validated if validated is not None else False,
                        "metas": metas,
                        "worker_version_id": self.worker_version_id,
                    }
                ]
                CachedEntity.insert_many(to_insert).execute()
            except IntegrityError as e:
                logger.warning(f"Couldn't save created entity in local cache: {e}")

        return entity["id"]

    def create_transcription_entity(
        self,
        transcription: str,
        entity: str,
        offset: int,
        length: int,
        confidence: Optional[float] = None,
    ) -> Optional[Dict[str, Union[str, int]]]:
        """
        Create a link between an existing entity and an existing transcription.
        If cache support is enabled, a `CachedTranscriptionEntity` will also be created.

        :param transcription: UUID of the existing transcription.
        :param entity: UUID of the existing entity.
        :param offset: Starting position of the entity in the transcription's text,
           as a 0-based index.
        :param length: Length of the entity in the transcription's text.
        :param confidence: Optional confidence score between 0 or 1.
        :returns: A dict as returned by the ``CreateTranscriptionEntity`` API endpoint,
           or None if the worker is in read-only mode.
        """
        assert transcription and isinstance(
            transcription, str
        ), "transcription shouldn't be null and should be of type str"
        assert entity and isinstance(
            entity, str
        ), "entity shouldn't be null and should be of type str"
        assert (
            offset is not None and isinstance(offset, int) and offset >= 0
        ), "offset shouldn't be null and should be a positive integer"
        assert (
            length is not None and isinstance(length, int) and length > 0
        ), "length shouldn't be null and should be a strictly positive integer"
        assert (
            confidence is None or isinstance(confidence, float) and 0 <= confidence <= 1
        ), "confidence should be null or a float in [0..1] range"
        if self.is_read_only:
            logger.warning(
                "Cannot create transcription entity as this worker is in read-only mode"
            )
            return

        body = {
            "entity": entity,
            "length": length,
            "offset": offset,
            "worker_version_id": self.worker_version_id,
        }
        if confidence is not None:
            body["confidence"] = confidence

        transcription_ent = self.request(
            "CreateTranscriptionEntity",
            id=transcription,
            body=body,
        )
        # TODO: Report transcription entity creation

        if self.use_cache:
            # Store transcription entity in local cache
            try:
                CachedTranscriptionEntity.create(
                    transcription=transcription,
                    entity=entity,
                    offset=offset,
                    length=length,
                    worker_version_id=self.worker_version_id,
                    confidence=confidence,
                )
            except IntegrityError as e:
                logger.warning(
                    f"Couldn't save created transcription entity in local cache: {e}"
                )
        return transcription_ent

    def list_transcription_entities(
        self,
        transcription: Transcription,
        worker_version: Optional[Union[str, bool]] = None,
    ):
        """
        List existing entities on a transcription
        This method does not support cache

        :param transcription: The transcription to list entities on.
        :param worker_version: Restrict to entities created by a worker version with this UUID. Set to False to look for manually created transcriptions.
        """
        query_params = {}
        assert transcription and isinstance(
            transcription, Transcription
        ), "transcription shouldn't be null and should be a Transcription"

        if worker_version is not None:
            assert isinstance(
                worker_version, (str, bool)
            ), "worker_version should be of type str or bool"

            if isinstance(worker_version, bool):
                assert (
                    worker_version is False
                ), "if of type bool, worker_version can only be set to False"
            query_params["worker_version"] = worker_version

        return self.api_client.paginate(
            "ListTranscriptionEntities", id=transcription.id, **query_params
        )

    def list_corpus_entities(
        self,
        name: Optional[str] = None,
        parent: Optional[Element] = None,
    ):
        """
        List all entities in the worker's corpus
        This method does not support cache
        :param name: Filter entities by part of their name (case-insensitive)
        :param parent Element: Restrict entities to those linked to all transcriptions of an element and all its descendants. Note that links to metadata are ignored.
        """
        query_params = {}

        if name is not None:
            assert name and isinstance(name, str), "name should be of type str"
            query_params["name"] = name

        if parent is not None:
            assert isinstance(parent, Element), "parent should be of type Element"
            query_params["parent"] = parent.id

        return self.api_client.paginate(
            "ListCorpusEntities", id=self.corpus_id, **query_params
        )



# -*- coding: utf-8 -*-
"""
General utility functions and classes.
"""

import datetime
from timeit import default_timer


class Timer(object):
    """
    A context manager to help measure execution times.

    Example
    ---
    ```
    with Timer() as t:
        # do something interesting
    print(t.delta)  # X days, X:XX:XX
    ```
    """

    def __init__(self):
        self.timer = default_timer

    def __enter__(self):
        self.start = self.timer()
        return self

    def __exit__(self, *args):
        end = self.timer()
        self.elapsed = end - self.start
        self.delta = datetime.timedelta(seconds=self.elapsed)
